#
# Sensus Consulting Ltd. (C) 1997-1998
# Matt Newman <matt@sensus.org>
#
# Abstract Class for Generic Tree
#
#INDEX\
proc tcl++::atree {} {}

package require tcl++

class tcl++::atree {
    public {
		constructor {args} {eval configure $args}

		method add {id args} {error Abstract}
		method delete {id} {error Abstract}
		method get {id args} {error Abstract}
		method put {id fld val args} {error Abstract}

		method children {id {order -increasing}} {error Abstract}
		method parent {id} {error Abstract}

		method lookup {path} {}
		method path {id args} {}
		method walk {idRef body} {}
		#
		# Sort by list of fields.
		# field may be pair, where first no is "asc" or "desc" for
		# ascending or decending sort order.
		method sort {list args} {}
		#
		# Convert subtree indicated by id into an XML data-stream.
		#
		method export {{id 0}}
		#
		# Populate this tree, below $id with the packed tree in $data
		#
		method import {data {id 0} {skip 1}}
		#
		# Debug & Convenience routines
		#
		method print {}

		public variable separator /
    }
    protected {
    	method sortPair {flds id1 id2} 

		variable uid 0
		variable store
    }
	#
	# Private methods
	#
	private method _lookup1 {id name}
}

body tcl++::atree::path {id args} {
	set name [get $id name]
	set pid [parent $id]

	if {$pid == $id} {
		return ""
	} else {
		set path [path $pid]
		if {$path == ""} {
			return $name
		} else {
			return "$path$separator$name"
		}
	}
}
body tcl++::atree::lookup {path} {
	set id 0
	set p [split $path $separator
	while {![lempty $p]} {
		set id [_lookup1 $id [lvarpop p]]
		if {$id == ""} {
			error "$path not found"
		}
	}
	return $id
}
body tcl++::atree::sort {list args} {
	return [lsort -command [code $this sortPair $args] $list]
}
body tcl++::atree::sortPair {flds id1 id2} {
	set m(asc) 1
	set m(desc) -1
	set dir asc
	foreach fld $flds {
		if {[llength $fld]>1} {
			set dir [lindex $fld 0]
			set fld [lindex $fld 1]
		}
		set v1 [get $id1 $fld ""]
		set v2 [get $id2 $fld ""]
		if {$v1 < $v2} {
			return [expr {$m($dir) * -1}]
		} elseif {$v1 > $v2} {
			return [expr {$m($dir) * 1}]
		}
	}
	return 0
}
body tcl++::atree::walk {idRef body} {
	upvar 1 $idRef id

	set stack [children 0 -decreasing]

	upvar 1 _error_ ret
	while {![lempty $stack]} {
		set id [lvarpop stack]
		set retCode [uplevel 1 [list catch $body _error_]]
		switch -- $retCode {
		0	{#TCL_OK}
		2	{#TCL_RETURN
			return -code return $ret
			}
		3	{#TCL_BREAK
			break
			}
		4	{#TCL_CONTINUE
			continue
			}
		1	-
		default	{#TCL_ERROR
			global errorInfo errorCode
			return -code $retCode \
				-errorinfo $errorInfo \
				-errorcode $errorCode $ret
			}
		}
		foreach cid [children $id -decreasing] {
			lvarpush stack $cid
		}
	}
}
#
# Convert subtree indicated by id into a data-stream.
#
body tcl++::atree::export {{id 0}} {
	set attrs [get $id]
	set tag [pget $attrs class]
	set attrs [pdel $attrs id]
	set attrs [pdel $attrs class]
	set attrs [eval xml::putattrs $attrs]

	set cids [children $id]
	if {[lempty $cids]} {
		set data "<${tag} $attrs />\n"
	} else {
		set data "<${tag} $attrs >\n"
		foreach cid $cids {
			append data [export $cid]
		}
		append data "</${tag}>\n"
	}
	return $data
}
#
# Populate this tree, below $id with the packed tree in $data
#
body tcl++::atree::import {data {id 0} {skip 1}} {
	set data [xml::parse $data]

	set n 0
	set stack {}
	foreach {tag close empty param text} $data {
		if {$tag == ""} continue
		if {$skip} {
			set skip 0
			continue
		}
		if {[string length $close]!=0} {
			set id [lvarpop stack]
			continue
		}
		set cmd [list add $id class $tag]
		foreach {opt val} [xml::getattrs $param] {
			lappend cmd $opt $val
		}
		set cid [eval $cmd]
		incr n
		if {[string length $empty]==0} {
			lvarpush stack $id
			set id $cid
		}
	}
	if {[llength $stack]!=0} {
		error "unmatched tags: stack=$stack"
	}
	return $n
}
#
# Debug & Convenience routines
#
body tcl++::atree::print {} {
	walk id {
		puts stdout [path $id]
		#update idletasks
	}
}
#
# Private methods
#
body tcl++::atree::_lookup1 {id name} {
	foreach cid [children $id] {
		set cname [get $cid name]
		if {[string compare $name $cname]==0} {
			return $cid
		}
	}
	return ""
}
